import logging
from argparse import Namespace
from pathlib import Path

from talisman_tools.configure import read_config
from talisman_tools.plugin import TrainerPlugins
from tp_interfaces.abstract import AbstractTrainer
from tp_interfaces.abstract.processor.trainer import DEFAULT_CACHE_DIR

logger = logging.getLogger(__name__)


def train(args: Namespace):
    config = read_config(args.config_path)
    trainer_plugin = config['plugin']
    trainer_model = config['model']
    trainer_config = config['config']
    if 'evaluation_strategy' in trainer_config.get('trainer', {}) and not trainer_config['trainer']['evaluation_strategy']:
        trainer_config['trainer']['evaluation_strategy'] = 'no'
    trainer: AbstractTrainer = TrainerPlugins.plugins[trainer_plugin][trainer_model].from_config(trainer_config)

    cache_dir = Path(args.cache_dir) if args.cache_dir is not None else DEFAULT_CACHE_DIR
    if args.ignore_cache:
        cache_dir = None

    train_results = trainer.train(cache_dir)
    train_results.model.save(Path(args.trained_model_path))



from argparse import ArgumentParser
from typing import Callable, Dict, Tuple


def configure_subparsers(
        parser: ArgumentParser,
        subparsers: Dict[str, Tuple[str, Callable[[ArgumentParser], None]]]
) -> None:
    subparsers_handler = parser.add_subparsers(help='Available sub-commands')
    for subparser_name, (description, configure) in subparsers.items():
        subparser = subparsers_handler.add_parser(subparser_name, help=description)
        try:
            configure(subparser)
        except ImportError:
            pass



from argparse import ArgumentParser
from typing import Callable, Dict, Tuple

from talisman_tools.commands.model.commands.check_performance import configure_check_performance_parser
from talisman_tools.commands.model.commands.configure import configure_configure_parser
from talisman_tools.commands.model.commands.evaluate import configure_evaluate_parser
from talisman_tools.commands.model.commands.process import configure_process_parser

# Note that subparsers should change the default `processor_action` argument
SUBPARSERS: Dict[str, Tuple[str, Callable[[ArgumentParser], None]]] = {
    'check_performance': ('Check performance of document processors', configure_check_performance_parser),
    'configure': ('Configure models from configuration files', configure_configure_parser),
    'evaluate': ('Evaluate document processors on a number of metrics', configure_evaluate_parser),
    'process': ('Process documents with document processor', configure_process_parser)
}



import logging
from argparse import Namespace

from fastapi import FastAPI

from talisman_tools.commands.servers.commands.loader.methods.load import register_load_docs
from talisman_tools.commands.servers.server_helper import async_register_context_manager, register_exception_handlers
from tp_interfaces.domain.manager import DomainManager
from tp_interfaces.loader.interfaces import Loader

logger = logging.getLogger(__name__)


def action(args: Namespace, loader: Loader) -> FastAPI:
    app = FastAPI(title='kb loader server', description='kb loader server')

    register_load_docs(
        app=app,
        loader=loader,
        logger=logger
    )

    register_exception_handlers(app)
    async_register_context_manager(app, DomainManager())
    async_register_context_manager(app, loader)

    return app



from argparse import ArgumentParser, Namespace
from typing import Callable

from fastapi import FastAPI

from talisman_tools.arguments.domain import get_domain_factory
from talisman_tools.arguments.processor import get_processor_factory
from talisman_tools.plugin.cli import add_extra_cli_arguments


def configure_processor_parser(parser: ArgumentParser) -> None:
    domain_factory = get_domain_factory(parser)
    processor_factory = get_processor_factory(parser)

    extra_actions = add_extra_cli_arguments(parser, 'processor')

    def get_action() -> Callable[[Namespace], FastAPI]:
        def action_with_extra(args: Namespace) -> FastAPI:
            for action in extra_actions:
                action(args)
            from .action import action
            domain_factory(args)
            return action(args, processor_factory)

        return action_with_extra

    parser.set_defaults(server_action=get_action)



from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.arguments.reader import get_reader_factory
from talisman_tools.parser_helper import configure_subparsers


def configure_dataset_parser(parser: ArgumentParser) -> None:
    from .commands import SUBPARSERS

    reader_factory = get_reader_factory(parser, {'docs_path': '<docs path>'})

    def get_action() -> Callable[[Namespace], None]:
        def action(args: Namespace) -> None:
            args.dataset_action()(reader_factory(args)['docs_path'], args)

        return action

    parser.set_defaults(action=get_action)
    configure_subparsers(parser, SUBPARSERS)



from dataclasses import dataclass
from functools import partial
from itertools import chain, zip_longest
from typing import Iterable, Iterator, Union

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractDomainType
from tdm.datamodel.facts import AtomValueFact, ConceptFact, PropertyFact, RelationFact
from typing_extensions import Self

from talisman_tools.commands.model.commands.evaluate.evaluation import _evaluate
from talisman_tools.commands.model.commands.evaluate.helpers import get_original_type_id
from tp_interfaces.helpers.datamodel.domain import get_id

NERFact = Union[ConceptFact, AtomValueFact]
LinkFact = Union[RelationFact, PropertyFact]


@dataclass(frozen=True, slots=True)
class RelationTupleView:
    """"
    A utility class that represents relation facts as dataclasses (without fact ids)
    """

    type_id: str | AbstractDomainType
    source: NERFact
    target: NERFact

    def __post_init__(self):
        """
        Replace domain type id of relation with the original one.
        """
        object.__setattr__(self, 'type_id', get_original_type_id(get_id(self.type_id)))

    @classmethod
    def from_fact(cls, fact: RelationFact) -> Self:
        return RelationTupleView(fact.type_id, fact.source, fact.target)

    @classmethod
    def from_doc(cls, doc: TalismanDocument, excluded_types: dict) -> Iterator[Self]:
        relation_blacklist = set(excluded_types.get('relations', []))

        for fact in doc.get_facts(LinkFact):
            if get_original_type_id(get_id(fact.type_id)) not in relation_blacklist:
                yield cls.from_fact(fact)


def _fact_type(query, fact: RelationTupleView) -> str:
    return fact.type_id


def _source_target_types(query, link_fact: RelationTupleView) -> tuple[str, str]:
    return get_id(link_fact.source.type_id), get_id(link_fact.target.type_id)


def _link_signature(query, link_fact: RelationTupleView) -> tuple[str, str, str]:
    return get_id(link_fact.source.type_id), get_id(link_fact.target.type_id), link_fact.type_id


_LINK_CATEGORIZERS = {
    "fact_type": _fact_type,
    "source_target_types": _source_target_types,
    "signature": _link_signature
}


def evaluate_relext(predicted: Iterable[TalismanDocument], gold: Iterable[TalismanDocument], eval_config: dict) -> dict[str, dict]:
    return _evaluate(
        predicted, gold, partial(RelationTupleView.from_doc, excluded_types=eval_config.get('excluded_types', {})), _LINK_CATEGORIZERS
    )


def evaluate_relext_upper_bound(
    predicted: Iterable[TalismanDocument],
    gold: Iterable[TalismanDocument],
    eval_config: dict
) -> dict[str, dict]:
    predicted, gold = tuple(predicted), tuple(gold)
    predicted_w_gold_rels = []

    for pred_doc, gold_doc in zip_longest(predicted, gold):
        if pred_doc is None or gold_doc is None:
            raise ValueError("Predicted and gold documents iterables length mismatch")

        pred_entity_facts = set(pred_doc.get_facts(NERFact))
        gold_entity_facts = set(gold_doc.get_facts(NERFact)).intersection(pred_entity_facts)

        def filter_(fact: RelationFact) -> bool:
            return fact.source in gold_entity_facts and fact.target in gold_entity_facts

        possible_to_find_gold_rels = gold_doc.get_facts(RelationFact, filter_=filter_)

        predicted_w_gold_rels.append(
            gold_doc.without_facts(gold_doc.get_facts()).with_facts(chain(gold_entity_facts, possible_to_find_gold_rels))
        )

    return evaluate_relext(predicted_w_gold_rels, gold, eval_config)



import logging
from logging import Logger

from fastapi import FastAPI, HTTPException, Request
from tdm import TalismanDocumentModel

from tp_interfaces.domain.manager import DomainManager
from tp_interfaces.loader.interfaces import Loader
from tp_interfaces.logging.context import update_log_extras

_logger = logging.getLogger(__name__)


def register_load_docs(
        app: FastAPI,
        loader: Loader,
        logger: Logger = _logger
):
    cfg_type = loader.config_type

    @app.post('/', response_model=str)
    async def load_docs(request: Request, *, message: TalismanDocumentModel, config: cfg_type):
        with update_log_extras(doc_id=message.id):
            async with DomainManager() as manager:
                await manager.domain
            doc = message.deserialize()
            try:
                status = await loader.load_doc_and_bind_facts(doc, config)
            except Exception as e:
                logger.error("Non handled exception while request processing", exc_info=e)
                raise HTTPException(status_code=500)

            if status is None:
                raise HTTPException(status_code=500, detail='No document was loaded')
            return status



import time
from functools import partial
from itertools import chain
from pathlib import Path
from typing import Optional

from tdm import TalismanDocument
from tdm.datamodel.domain import Domain
from tdm.datamodel.facts import AtomValueFact, ConceptFact, MentionFact, PropertyFact

from talisman_tools.commands.model.commands.evaluate.evaluators import evaluators
from talisman_tools.commands.model.commands.evaluate.helpers import build_domain, filter_facts, print_scores, replace_with_domain
from talisman_tools.configure.configure import read_config
from talisman_tools.plugin import DomainPlugins
from tp_interfaces.abstract import AbstractDocumentProcessor
from tp_interfaces.domain.manager import DomainManager
from tp_interfaces.readers.abstract import AbstractReader


def keep_nerc(doc: TalismanDocument) -> TalismanDocument:

    facts = chain(
        doc.get_facts(ConceptFact),
        doc.get_facts(AtomValueFact),
        doc.get_facts(MentionFact),
        doc.get_facts(PropertyFact)
    )
    return doc.without_facts(doc.get_facts()).with_facts(facts)


def clear_values(doc: TalismanDocument) -> TalismanDocument:
    return doc.with_facts(
        [f.with_changes(value=tuple()) for f in doc.get_facts(ConceptFact)],  # TODO: replace value with None instead of empty tuple
    )


def clear_chains(doc: TalismanDocument) -> TalismanDocument:
    pass


mode = {
    'all': lambda doc: doc.without_facts(doc.get_facts()),  # start from clear document (no facts provided)
    'nerc': lambda doc: doc.without_facts(doc.get_facts()),  # start from clear document (no facts provided)
    'relext': keep_nerc,  # start from document with concept, value, mention and property facts (no link facts, no fact values)
    'dmb': clear_values,  # start from document with facts without values
    'coref': clear_chains
}


async def evaluate(
        processor: AbstractDocumentProcessor,
        eval_mode: str,
        reader: AbstractReader,
        config_path: Path,
        eval_config_path: Path,
        input_reader: Optional[AbstractReader] = None
):
    """
    Core function that evaluates a document processor under the given task.
    """

    # load gold documents
    gold_docs = tuple(reader.read())

    # build domain based on the gold documents
    domain: Domain = build_domain(gold_docs)
    DomainManager().set_producer(DomainPlugins.plugins[None]()(domain))  # DMBCommonnessStub

    # replace old type_id of fact with domain types
    replace_with_domain_ = partial(replace_with_domain, domain=domain)
    gold_docs = tuple(map(replace_with_domain_, gold_docs))

    # load raw documents (to be processed)
    actual_docs = tuple(map(mode[eval_mode], gold_docs)) if input_reader is None else tuple(map(replace_with_domain_, input_reader.read()))

    # process raw documents
    async with DomainManager() and processor:

        evaluation_start = time.time()

        processor_config_type = processor.config_type
        config = processor_config_type.model_validate(read_config(config_path)) if config_path else processor_config_type()
        actual_docs = await processor.process_docs(actual_docs, config)

        evaluation_end = time.time()

    # evaluate result
    filtered_docs = [filter_facts(doc, domain) for doc in actual_docs]
    eval_config = read_config(eval_config_path) if eval_config_path else {}
    scores = {name: evaluator(filtered_docs, gold_docs, eval_config.get(name, {})) for name, evaluator in evaluators[eval_mode].items()}

    print_scores(scores)
    print(f'Total evaluation time: {evaluation_end - evaluation_start:.04f} seconds')



import json
from dataclasses import replace
from itertools import chain
from typing import TypeVar, Union

from more_itertools.more import always_iterable
from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractDomainType, AbstractFact, AbstractLinkFact, AbstractValue
from tdm.datamodel.domain import Domain
from tdm.datamodel.facts import AtomValueFact, ConceptFact, MentionFact, PropertyFact, RelationFact

from tp_interfaces.domain.abstract import RelExtModel
from tp_interfaces.domain.model.types import AtomValueType, ConceptType, IdentifyingPropertyType, PropertyType, RelationType
from tp_interfaces.helpers.datamodel.domain import get_id

DELIMITER = '::'


def unique_relation_type_id(type_id: str, source_type_id: str, target_type_id: str):
    """
    In some datasets (e.g. DocRED) relations can be between different sets of source and target entity types. To meet requirements of
    domain we construct unique relation type ids for them (e.g. rel.type_id=f'{rel.type_id}::{source.type_id}_{target.type_id}').
    """
    return (type_id + DELIMITER + source_type_id + '_' + target_type_id).replace(' ', '_')


def get_original_type_id(type_id: str):
    """
    In some datasets (e.g. DocRED) relations can be between different sets of source and target entity types. To meet requirements of
    domain we construct unique relation type ids for them (e.g. rel.type_id=f'{rel.type_id}::{source.type_id}_{target.type_id}').
    This function returns original relation type id.
    """
    return type_id.split(DELIMITER)[0]


def print_scores(scores: dict[str, dict]):
    def round_floats(val, precision=4):
        if isinstance(val, int):
            return val
        if isinstance(val, float):
            return round(val, precision)
        if isinstance(val, dict):
            return {k: round_floats(v) for k, v in val.items()}
        raise ValueError

    def stringify_keys(d: dict):
        ret = {}
        for key, val in d.items():
            if isinstance(key, (tuple, frozenset)):
                key = str(key)
            if isinstance(val, dict):
                val = stringify_keys(val)

            ret[key] = val

        return ret

    json_repr = json.dumps(stringify_keys(round_floats(scores)), sort_keys=True, indent=2)
    print(json_repr)


_Value = TypeVar('_Value', bound=AbstractValue)


def _get_value_type(value_fact: AtomValueFact) -> _Value:
    value_types = list({v.__class__ for v in always_iterable(value_fact.value)})
    if len(value_types) != 1 or not isinstance(value_types[0], AbstractValue):
        return AbstractValue

    return value_types[0]


def get_domain_type(fact: AbstractFact, pretrained_types: dict[str, str] | None = None,
                    nerc_pretrained: bool = True, type_names: dict[str, str] = None) -> AbstractDomainType:
    """
    Returns fact's domain type (or build fake one).
    NOTE: each type must have a unique id.
    """

    if hasattr(fact, 'type_id') and isinstance(fact.type_id, AbstractDomainType):
        return fact.type_id

    pretrained = pretrained_types.get(fact.type_id) if hasattr(fact, 'type_id') else None
    type_name = type_names.get(fact.type_id) if hasattr(fact, 'type_id') and type_names else None

    if isinstance(fact, ConceptFact):
        return ConceptType(id=fact.type_id, name=type_name)
    if isinstance(fact, AtomValueFact):
        value_type = _get_value_type(fact)
        return AtomValueType(fact.type_id, value_type, id=fact.type_id, _pretrained_nerc_models=(pretrained,) if pretrained else ())
    if isinstance(fact, MentionFact):
        return get_domain_type(fact.value, pretrained_types=pretrained_types, nerc_pretrained=nerc_pretrained, type_names=type_names)

    if not isinstance(fact, AbstractLinkFact):
        raise ValueError

    source_type = get_domain_type(fact.source, pretrained_types=pretrained_types, nerc_pretrained=nerc_pretrained, type_names=type_names)
    target_type = get_domain_type(fact.target, pretrained_types=pretrained_types, nerc_pretrained=nerc_pretrained, type_names=type_names)

    # Now fact is property fact or relation fact
    # In some datasets (e.g. DocRED) relations can be between different sets of source and target entity types.
    # To meet requirements of domain we construct unique relation type ids for them
    params = {
        'name': type_name,
        'source': source_type,
        'target': target_type,
        'id': unique_relation_type_id(fact.type_id, get_id(source_type), get_id(target_type))
    }

    if isinstance(fact, PropertyFact) and nerc_pretrained:
        params['_pretrained_nerc_models'] = (pretrained,) if pretrained else ()
        return IdentifyingPropertyType(**params)

    params['_pretrained_relext_models'] = (RelExtModel(relation_type=pretrained),) if pretrained else ()

    if isinstance(fact, PropertyFact):
        return PropertyType(**params)
    if isinstance(fact, RelationFact):
        return RelationType(**params)
    raise ValueError


def build_domain(gold_docs: tuple[TalismanDocument, ...], pretrained_types: dict[str, str] | None = None, nerc_pretrained: bool = True,
                 type_names: dict[str, str] | None = None) -> Domain:
    """
    Builds domain based on the facts from gold documents.
    In the case if type_id of some fact is string, fake AbstractDomainType is created, otherwise original type_id is used.
    """
    return Domain(
        get_domain_type(fact, pretrained_types, nerc_pretrained, type_names) for fact in chain.from_iterable(
            doc.get_facts() for doc in gold_docs
        )
    )


def replace_with_domain(doc: TalismanDocument, domain: Domain, pretrained_types: dict[str, str] | None = None,
                        nerc_pretrained: bool = True, type_names: dict[str, str] | None = None):
    """
    Creates a new facts with domain type instead of its identifier.
    """

    def _replace_with_domain(fact: AbstractFact):
        if isinstance(fact, Union[PropertyFact | RelationFact]):
            fact = replace(fact, type_id=get_domain_type(fact, pretrained_types, nerc_pretrained, type_names))
        return fact.replace_with_domain(domain)

    facts = map(_replace_with_domain, doc.get_facts())
    return doc.without_facts(doc.get_facts()).with_facts(facts=facts, update=True)


def filter_facts(doc: TalismanDocument, domain: Domain):
    """
    Drops facts from document that are not represented in the domain.
    """

    def _filter_facts(fact: AbstractFact):
        try:
            if hasattr(fact, 'type_id'):
                domain.get_type(get_id(fact.type_id))
        except KeyError:
            return False
        return True

    facts = doc.get_facts(filter_=_filter_facts)
    return doc.without_facts(doc.get_facts()).with_facts(facts, update=True)



from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.plugin import KBPlugins
from tp_interfaces.helpers.io import read_json
from tp_interfaces.loader.interfaces import Loader


def get_loader_factory(parser: ArgumentParser) -> Callable[[Namespace], Loader]:
    parser.add_argument('-loader_config_path', metavar='<kb config path>', required=True)

    def load_kb(args: Namespace) -> Loader:
        config = read_json(args.loader_config_path)
        kb = KBPlugins.plugins[config.get("plugin")][config.get("model", "default")].from_config(config.get('config', {}))
        return kb

    return load_kb



import asyncio
import logging
from functools import partial
from logging import Logger
from operator import itemgetter
from typing import Any, List, Optional, Sequence, Tuple, Union

from fastapi import FastAPI, HTTPException, Request
from tdm import TalismanDocumentModel

from talisman_tools.commands.servers.server_helper import log_debug_data
from tp_interfaces.abstract import AbstractDocumentProcessor, ImmutableBaseModel
from tp_interfaces.abstract.exception import WrongRequest
from tp_interfaces.helpers.batch import AbstractModelInput, async_batch_process_inputs
from tp_interfaces.logging.context import update_log_extras
from tp_interfaces.logging.time import TimeMeasurer

_logger = logging.getLogger(__name__)


def register_process_docs(
        app: FastAPI,
        endpoint: str,
        processor: AbstractDocumentProcessor,
        logger: Logger = _logger
):
    config_model = processor.config_type

    class ModelInput(AbstractModelInput):
        message: TalismanDocumentModel
        config: config_model

        def get_message(self) -> Any:
            return self.message

        def get_config(self) -> Optional[ImmutableBaseModel]:
            return self.config

    async def process_with_config(messages: Sequence[TalismanDocumentModel], config: ImmutableBaseModel) \
            -> Tuple[TalismanDocumentModel, ...]:
        docs = tuple(message.deserialize() for message in messages)
        with update_log_extras(doc_id=[doc.id for doc in docs]):
            output_docs = await processor.process_docs(docs, config)
            return tuple(TalismanDocumentModel.serialize(output_doc) for output_doc in output_docs)

    @app.post(endpoint, response_model=Union[List[TalismanDocumentModel], TalismanDocumentModel], response_model_exclude_none=True)
    async def process(request: Request, *, messages: Union[List[ModelInput], ModelInput]):
        response_post_processor = lambda x: x
        if not isinstance(messages, list):
            messages = [messages]
            response_post_processor = itemgetter(0)

        with (update_log_extras(message_id=[message.message.id for message in messages])):
            log_debug_data(logger, f"got {len(messages)} documents for processing", request)

            task_wrapper = partial(async_batch_process_inputs, inputs=messages, processor=process_with_config)
            task = asyncio.create_task(task_wrapper())

            try:
                while not task.done():
                    if await request.is_disconnected():
                        task.cancel()
                        logger.error("Client disconnected.")
                        raise HTTPException(status_code=499, detail="Client disconnected.")
                    await asyncio.sleep(0.1)  # I think we should not check disconnection more frequently
                response = await task
            except asyncio.CancelledError as e:
                logger.error("Document processing was cancelled", exc_info=e)
                raise HTTPException(status_code=499, detail=format_error_response(e))
            except WrongRequest as e:
                logger.error("WrongRequest exception while processing request", exc_info=e)
                raise HTTPException(status_code=422, detail=format_error_response(e))
            except Exception as e:
                logger.error("Exception while processing request", exc_info=e)
                raise HTTPException(status_code=500)
            log_debug_data(logger, f"processed {len(response)} documents", request)

            with TimeMeasurer("response post processing", inline_time=True, logger=_logger):
                return response_post_processor(response)


def format_error_response(e: BaseException) -> str:
    return str(e) if len(str(e)) else "Internal error: " + e.__repr__()



import logging.config
from contextlib import AbstractContextManager
from typing import AsyncContextManager

from fastapi import FastAPI, Request
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import RequestValidationError
from starlette import status
from starlette.responses import JSONResponse

_logger = logging.getLogger(__name__)


def register_context_manager(app: FastAPI, context_manager: AbstractContextManager):
    @app.on_event("startup")
    async def enter_manager():
        context_manager.__enter__()

    @app.on_event("shutdown")
    async def exit_manager():
        context_manager.__exit__(None, None, None)


def async_register_context_manager(app: FastAPI, context_manager: AsyncContextManager):
    @app.on_event("startup")
    async def enter_manager():
        await context_manager.__aenter__()

    @app.on_event("shutdown")
    async def exit_manager():
        await context_manager.__aexit__(None, None, None)


def register_exception_handlers(app: FastAPI, logger: logging.Logger = _logger):
    @app.exception_handler(RequestValidationError)
    async def validation_exception_handler(request, exc: RequestValidationError):
        error_content = jsonable_encoder({"detail": exc.args})
        logger.error("Pydantic validation error", extra=error_content)
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content=error_content,
        )


def log_debug_data(logger: logging.Logger, msg: str, request: Request = None, **kwargs) -> None:
    extras = {}
    if request is not None:
        extras['client'] = request.client
    extras.update({key: jsonable_encoder(value) for key, value in kwargs.items()})
    logger.debug(msg, extra=extras)



import logging
from argparse import Namespace

from fastapi import FastAPI

from talisman_tools.commands.servers.commands.converter.methods.convert import register_convert_doc
from talisman_tools.commands.servers.server_helper import async_register_context_manager, register_exception_handlers
from tp_interfaces.domain.manager import DomainManager

logger = logging.getLogger(__name__)


def action(args: Namespace) -> FastAPI:
    app = FastAPI(title='document converter service')

    register_convert_doc(
        app=app,
        logger=logger
    )

    register_exception_handlers(app, logger)
    async_register_context_manager(app, DomainManager())
    return app



from argparse import ArgumentParser
from typing import Callable, Dict, Tuple

from talisman_tools.commands.dataset.commands.convert import configure_convert_parser

# Note that subparsers should change the default `dataset_action` argument
SUBPARSERS: Dict[str, Tuple[str, Callable[[ArgumentParser], None]]] = {
    'convert': ('', configure_convert_parser)
}



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable, Iterable

from tdm import TalismanDocument

from talisman_tools.plugin import SerializerPlugins


def get_serializer_factory(parser: ArgumentParser):
    serializers = SerializerPlugins.flattened
    argument_group = parser.add_argument_group(title="Output documents arguments")
    argument_group.add_argument('output', type=Path, metavar='<output path>')
    argument_group.add_argument('-serializer', type=str, metavar='<serializer type>', choices=set(serializers.keys()), default='default')

    def get_serializer(args: Namespace) -> Callable[[Iterable[TalismanDocument]], None]:
        from functools import partial
        return partial(serializers[args.serializer]().serialize, path=args.output)

    return get_serializer



from argparse import ArgumentParser, Namespace
from typing import Callable, Set

from .abstract import AbstractPluginManager


class CLIPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'CLI_OPTIONS')

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, dict) and all(CLIPluginManager._check_pair(k, v) for k, v in value.items())

    @staticmethod
    def _check_pair(key, value) -> bool:
        return isinstance(key, str) and isinstance(value, set) and all(isinstance(v, Callable) for v in value)  # TODO: improve validation


def add_extra_cli_arguments(parser: ArgumentParser, key: str) -> Set[Callable[[Namespace], None]]:
    from talisman_tools.plugin import CLIPlugins
    actions = set()
    for cli_factory in CLIPlugins.plugins.values():
        factories = cli_factory.get(key)
        if factories:
            actions.update(factory(parser) for factory in factories)
    return actions



from multiprocessing.pool import ApplyResult
from typing import Callable, Dict, TypeVar

from anyio import Semaphore
from starlette.concurrency import run_in_threadpool

from talisman_tools.helper.env import get_max_concurrent_threads, get_max_treated_jobs

MAX_CONCURRENT_THREADS = get_max_concurrent_threads()
MAX_THREADS_GUARD = Semaphore(MAX_CONCURRENT_THREADS)
_AnyResult = TypeVar('_AnyResult')


async def execute_concurrently(io_bound_fn: Callable[[], _AnyResult]) -> _AnyResult:
    # too many threads may degrade the performance
    async with MAX_THREADS_GUARD:
        return await run_in_threadpool(io_bound_fn)


def check_on_jobs(decoding_jobs: Dict[int, ApplyResult[_AnyResult]], *, ignore_max_treated_jobs: bool = False) -> Dict[int, _AnyResult]:
    completed = []
    results = {}
    for job_id, job in decoding_jobs.items():
        if job.ready():
            results[job_id] = job.get()
            completed.append(job_id)

        if not ignore_max_treated_jobs and len(completed) >= get_max_treated_jobs():
            break

    for job_id in completed:
        decoding_jobs.pop(job_id)

    return results



from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.arguments.domain import get_domain_factory
from tp_interfaces.abstract.processor.trainer import DEFAULT_CACHE_DIR


def configure_train_parser(parser: ArgumentParser) -> None:
    domain_factory = get_domain_factory(parser)

    parser.add_argument('config_path', type=str, metavar='<config.json>', help='Path to training configuration file')
    parser.add_argument('trained_model_path', type=str, metavar='<trained model path>', help='Path to save trained model to')
    parser.add_argument('-cache_dir', type=str, default=DEFAULT_CACHE_DIR, help='Path to directory with cache files')
    parser.add_argument('--ignore_cache', action='store_true', help='Do not use cache')

    def get_action() -> Callable[[Namespace], None]:
        def action(args: Namespace) -> None:
            from .action import train
            domain_factory(args)
            train(args)

        return action

    parser.set_defaults(action=get_action)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable

from tp_interfaces.abstract import AbstractDocumentProcessor


def configure_configure_parser(parser: ArgumentParser) -> None:
    parser.add_argument('out_path', type=Path, metavar='<model out path>')

    def get_action() -> Callable[[AbstractDocumentProcessor, Namespace], None]:
        def action(processor: AbstractDocumentProcessor, args: Namespace) -> None:
            processor.save(args.out_path)
            print(f"Saved {processor} model")

        return action

    parser.set_defaults(processor_action=get_action)



import logging

from talisman_tools.configure import configure_model
from tp_interfaces.abstract import AbstractDocumentProcessor

logger = logging.getLogger(__name__)


def wrap_model_action(config: dict) -> AbstractDocumentProcessor:
    """
    create model from config and wraps it with respect to specified wrapper and wrapper configuration
    :param config: dict with required "wrapper" and "to_wrap" keys. Optional "plugin" and "config" keys are taken into account
    :return: wrapped document processor
    """
    from talisman_tools.plugin import WrapperPlugins  # inline import to avoid circular imports
    if isinstance(config["wrapper"], dict):
        plugin = config["wrapper"].get("plugin")
        model = config["wrapper"]["model"]
    elif isinstance(config["wrapper"], str):
        plugin = None
        model = config["wrapper"]
    else:
        logger.error(f'"wrapper" configuration should be either `dict` or `str`. Actual type: {type(config["wrapper"])}')
        raise ValueError
    try:
        plugin_models = WrapperPlugins.plugins[plugin]
    except KeyError as e:
        logger.error(f"'{plugin}' has no registered wrappers. Available plugins: {list(WrapperPlugins.plugins.keys())}", exc_info=e)
        raise
    try:
        wrapper_factory = plugin_models[model]
    except KeyError as e:
        logger.error(f"'{plugin}' contains no '{model}' wrapper. Available models: {list(plugin_models.keys())}", exc_info=e)
        raise
    try:
        return wrapper_factory.from_config(configure_model(config['to_wrap']), config.get("config", {}))
    except Exception as e:
        raise ValueError(f"Couldn't instantiate {plugin}.{model} with specified config") from e


def configure_model_action(config: dict) -> AbstractDocumentProcessor:
    """
    create model from config
    :param config: dict with required "model" key
    :return: configured model
    """
    return configure_model(config)



from argparse import ArgumentParser

from talisman_tools.helper.multiprocessing import setup_multiprocessing
from talisman_tools.parser_helper import configure_subparsers
from talisman_tools.talisman_logging import get_logging_factory


def main():
    parser = ArgumentParser(description='Talisman tools')
    logging_factory = get_logging_factory(parser)

    from talisman_tools.commands import SUBPARSERS
    configure_subparsers(parser, SUBPARSERS)

    args = parser.parse_args()
    logging_factory(args)  # initialize logging before performing any actions

    if not hasattr(args, 'action'):
        parser.print_help()
        exit(1)

    action = args.action()
    action(args)  # pass control to relevant subparser


if __name__ == "__main__":
    setup_multiprocessing()
    main()



from argparse import ArgumentParser
from typing import Callable, Dict, Tuple

from talisman_tools.commands.dataset.parser import configure_dataset_parser
from talisman_tools.commands.model.parser import configure_model_parser
from talisman_tools.commands.servers.parser import configure_server_parser
from talisman_tools.commands.train.parser import configure_train_parser

# Note that subparsers should change the default `action` argument
SUBPARSERS: Dict[str, Tuple[str, Callable[[ArgumentParser], None]]] = {
    'server': ('Configure server', configure_server_parser),
    'model': ('Evaluate and configure models', configure_model_parser),
    'dataset': ('Handle datasets', configure_dataset_parser),
    'train': ('Train document processors', configure_train_parser)
}



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable, Dict

from talisman_tools.plugin import ReaderPlugins
from tp_interfaces.readers.abstract import AbstractReader


def get_reader_factory(parser: ArgumentParser, arguments: Dict[str, str]) -> Callable[[Namespace], Dict[str, AbstractReader]]:
    readers = ReaderPlugins.flattened
    reader_choices = set(readers.keys())
    argument_group = parser.add_argument_group(title="Input documents arguments")
    argument_group.add_argument('reader', choices=reader_choices, help=f'Type of documents reader.')
    for arg, metavar in arguments.items():
        argument_group.add_argument(arg, type=Path, metavar=metavar)

    def get_reader(args: Namespace) -> Dict[str, AbstractReader]:
        reader = readers[args.reader]
        result = {}
        for arg in arguments:
            p = getattr(args, arg.lstrip('-'))
            if p is not None:
                result[arg] = reader(p)
        return result

    return get_reader



from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.arguments.processor import get_processor_factory
from talisman_tools.parser_helper import configure_subparsers


def configure_model_parser(parser: ArgumentParser) -> None:
    from .commands import SUBPARSERS

    processor_factory = get_processor_factory(parser)

    def get_action() -> Callable[[Namespace], None]:
        def action(args: Namespace):
            args.processor_action()(processor_factory(args), args)

        return action

    parser.set_defaults(action=get_action)

    configure_subparsers(parser, SUBPARSERS)



from tp_interfaces.abstract import AbstractDocumentProcessor
from .configure import wrap_model_from_config


def wrap_model(model: AbstractDocumentProcessor, config: dict, *, merge: bool = True):
    from talisman_tools.plugin import WrapperActionsPlugins
    WrapperActionsPlugins.add_wrapper_action(frozenset(('wrapped',)), lambda c: model)
    return wrap_model_from_config(config, merge=merge)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Tuple

from tp_interfaces.abstract import AbstractDocumentProcessor


def _named_path(arg: str) -> Tuple[str, Path]:
    if ":" not in arg:
        return "", Path(arg)
    name, path = arg.split(":", 1)
    return name, Path(path)


def get_processor_factory(parser: ArgumentParser, prefix: str = ''):
    argument_group = parser.add_argument_group(title='Model arguments')
    model_paths, wrapper, omit_merging = f'{prefix}model_paths', f'{prefix}wrapper', f'{prefix}omit_merging'

    argument_group.add_argument(
        f'-{prefix}m', f'--{model_paths}', action='append', type=_named_path, metavar='<model name:model path>', required=True,
        help='Named paths to configuration files or serialized models'
    )
    argument_group.add_argument(
        f'--{wrapper}', type=Path, metavar='<wrapper path>', help='Path to wrapper configuration file', default=None
    )
    argument_group.add_argument(f'--{omit_merging}', action='store_true', help='Disable model merging during configuration')

    def get_processor(args: Namespace) -> AbstractDocumentProcessor:
        from talisman_tools.configure import load_or_configure, wrap_model
        from talisman_tools.configure.configure import read_config

        processor = load_or_configure(model_or_config_named_paths=getattr(args, model_paths), merge=not getattr(args, omit_merging))
        if getattr(args, wrapper) is not None:
            processor = wrap_model(processor, read_config(getattr(args, wrapper)), merge=not getattr(args, omit_merging))
        return processor

    return get_processor



import os

MULTIPROCESSING_SETUP = False


def setup_multiprocessing():
    """Conventional forking is not safe in multithreaded environment. Use this function in __main__ to set up thread-safe version
    of spawning new processes.

    See https://docs.python.org/3/library/multiprocessing.html#contexts-and-start-methods
    """
    global MULTIPROCESSING_SETUP
    if MULTIPROCESSING_SETUP:
        return

    try:
        import torch.multiprocessing as multiprocessing
    except ImportError:
        import multiprocessing

    multiprocessing.set_start_method('spawn' if os.name == 'nt' else 'forkserver', force=True)
    MULTIPROCESSING_SETUP = True



from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.arguments.serializer import get_serializer_factory
from tp_interfaces.readers.abstract import AbstractReader


def configure_convert_parser(parser: ArgumentParser) -> None:
    serializer_factory = get_serializer_factory(parser)

    def get_action() -> Callable[[AbstractReader, Namespace], None]:
        def action(reader: AbstractReader, args: Namespace) -> None:
            serializer_factory(args)(reader.read())

        return action

    parser.set_defaults(dataset_action=get_action)



import logging
from logging import Logger
from typing import Any

from fastapi import FastAPI
from pydantic.main import BaseModel

from tp_interfaces.abstract import AbstractDocumentProcessor

_logger = logging.getLogger(__name__)


def register_schema(
        app: FastAPI,
        endpoint: str,
        processor: AbstractDocumentProcessor,
        logger: Logger = _logger
):
    config_model: BaseModel = processor.config_type

    @app.get(endpoint, response_model=dict[str, Any])
    async def schema() -> dict[str, Any]:
        return config_model.model_json_schema()



__all__ = [
    'ConfigurableReaderPlugins', 'CLIPlugins', 'DomainPlugins', 'EndpointPlugins', 'KBPlugins', 'ProcessorPlugins', 'ReaderPlugins',
    'SerializerPlugins', 'TrainerPlugins', 'WrapperPlugins', 'WrapperActionsPlugins'
]


from .cli import CLIPluginManager
from .domain import DomainPluginManager
from .endpoint import EndpointPluginManager
from .kb import KBPluginManager
from .processor import ProcessorPluginManager
from .reader import ConfigurableReaderPluginManager, ReaderPluginManager
from .serializer import SerializerPluginManager
from .trainer import TrainerPluginManager
from .wrapper import WrapperPluginManager
from .wrapper_actions import WrapperActionsPluginManager

CLIPlugins = CLIPluginManager()
ConfigurableReaderPlugins = ConfigurableReaderPluginManager()
DomainPlugins = DomainPluginManager()
EndpointPlugins = EndpointPluginManager()
KBPlugins = KBPluginManager()
ProcessorPlugins = ProcessorPluginManager()
ReaderPlugins = ReaderPluginManager()
SerializerPlugins = SerializerPluginManager()
TrainerPlugins = TrainerPluginManager()
WrapperPlugins = WrapperPluginManager()
WrapperActionsPlugins = WrapperActionsPluginManager()



__all__ = [
    'configure_model', 'load_or_configure', 'read_config', 'wrap_model_from_config',
    'wrap_model'
]

from .configure import configure_model, load_or_configure, read_config, wrap_model_from_config
from .wrap import wrap_model



import logging
import os
from logging import Logger
from typing import Optional

from fastapi import FastAPI, Request
from tdm import TalismanDocumentModel
from tdm.datamodel.domain import DocumentType, PropertyType
from tdm.v0 import TalismanDocumentModel as LegacyTalismanDocumentModel

from tp_interfaces.abstract import ImmutableBaseModel
from tp_interfaces.domain.manager import DomainManager
from tp_interfaces.helpers.batch import AbstractModelInput

_logger = logging.getLogger(__name__)

_doc_type_name = os.getenv('DOC_TYPE_NAME', 'Документ')

_metadata_map = {
    'title': 'Заголовок',
    'file_name': 'Название файла',
    'file_type': 'Формат файла',
    'size': 'Размер файла',
    'created_time': 'Дата создания',
    'access_time': 'Дата последнего просмотра',
    'modifired_time': 'Дата изменения',
    'publication_date': 'Дата публикации',
    'publication_author': 'Автор публикации',
    'description': 'Примечание',
    'url': 'Источник',
    'access_level': 'Уровень доступа',
    'user': 'Автор документа',
    'path': 'Путь до файла',
    'trust_level': 'Уровень доверия',
    'markers': 'Метки',
    'preview_text': 'Предпросмотр текста',
    'platform': 'Платформа',
    'account': 'Аккаунт',
    'story': 'Сюжет',
    'language': 'Язык документа',
    'job_id': 'Запуск сбора',
    'periodic_job_id': 'Периодический запуск сбора',
    'task_id': 'Задача импорта',
    'periodic_task_id': 'Периодическая задача импорта',
    'parent_uuid': 'Родительский документ'
}


def register_convert_doc(
        app: FastAPI,
        logger: Logger = _logger
):
    class ModelInput(AbstractModelInput):
        message: LegacyTalismanDocumentModel
        config: ImmutableBaseModel

        def get_message(self) -> LegacyTalismanDocumentModel:
            return self.message

        def get_config(self) -> Optional[ImmutableBaseModel]:
            return self.config

    @app.post('/', response_model=TalismanDocumentModel, response_model_exclude_none=True)
    async def convert(request: Request, *, message: ModelInput):
        logger.info("Recieved document: ", extra={"request": message.message})
        async with DomainManager() as manager:
            domain = await manager.domain

        titles = tuple(domain.get_types(PropertyType, filter_=lambda p: p.name == 'Название'))
        doc_type = next(domain.get_types(DocumentType, filter_=lambda t: t.name == _doc_type_name)).id

        message.message.set_domain(domain)
        document = message.message.to_doc(doc_type, titles, _metadata_map)
        logger.debug(f"document {document.id} successfully converted")
        result = TalismanDocumentModel.serialize(document)
        logger.info("Recieved document/converted document: ", extra={"request": message.message, "response": result})
        return result



import logging
from logging import Logger
from typing import Type

from fastapi import FastAPI, HTTPException, Request

from talisman_tools.commands.servers.server_helper import log_debug_data
from tp_interfaces.abstract import AbstractUpdatableModel, AbstractUpdate

_logger = logging.getLogger(__name__)


def register_update(app: FastAPI, model: AbstractUpdatableModel, input_type: Type[AbstractUpdate], logger: Logger = _logger):
    @app.post("/update/")
    async def update(request: Request, *, update_model: input_type):
        log_debug_data(logger, "update requested", request, update=update_model.model_dump_json())
        try:
            model.update(update_model)
            log_debug_data(logger, "successfull update", request, update=update_model.model_dump_json())
        except Exception as e:
            logger.error("error during update", exc_info=e, extra={"update": update_model.model_dump_json()})
            raise HTTPException(status_code=400, detail=str(e))



from collections import defaultdict
from itertools import zip_longest
from typing import Any, Callable, Iterable

from tdm import TalismanDocument

from .metrics import ir_categorized_queries_score, ir_macro_scores, ir_micro_scores


def _scores2dict(precision: float, recall: float, f1: float):
    return {"precision": float(precision), "recall": float(recall), "f1": float(f1)}


def _with_prefix(d: dict[str, Any], prefix: str) -> dict:
    return {f"{prefix}_{k}": v for k, v in d.items()}


def _evaluate(
        predicted: Iterable[TalismanDocument],
        gold: Iterable[TalismanDocument],
        objects_getter: Callable[[TalismanDocument], Iterable[Any]],
        categorizers: dict[str, Callable]
) -> dict[str, dict]:

    queries = []
    predicted_objects, gold_objects = [], []

    for pred_doc, gold_doc in zip_longest(predicted, gold):

        if pred_doc is None or gold_doc is None:
            raise ValueError('Predicted and gold documents iterables length mismatch')

        if pred_doc.id != gold_doc.id:
            raise ValueError(f'The document sequences are not identical {pred_doc.id} != {gold_doc.id}')

        if pred_doc.main_root.content != gold_doc.main_root.content:
            raise ValueError('Predicted and gold documents have different texts')

        predicted_objects.append(set(objects_getter(pred_doc)))
        gold_objects.append(set(objects_getter(gold_doc)))
        queries.append(gold_doc.main_root.content)

    macro_scores = _scores2dict(*ir_macro_scores(predicted_objects, gold_objects))
    micro_scores = _scores2dict(*ir_micro_scores(predicted_objects, gold_objects))
    final_scores = {**_with_prefix(macro_scores, 'macro'), **_with_prefix(micro_scores, 'micro'), 'categories': {}}

    for categorizer_name, categorizer in categorizers.items():
        category_scores = defaultdict(dict)

        for strategy_name, strategy in [('micro', ir_micro_scores), ('macro', ir_macro_scores)]:
            scores: dict[str, tuple[float, float, float]] = ir_categorized_queries_score(
                queries, predicted_objects, gold_objects, categorizer, strategy)
            scores: dict[str, dict[str, float]] = {k: _with_prefix(_scores2dict(*vals), strategy_name) for k, vals in scores.items()}
            for key, val in scores.items():
                category_scores[key].update(val)

        final_scores['categories'][categorizer_name] = dict(category_scores)

    return final_scores


def evaluate_categorized_objects(predicted_objects: dict[str, list[set]], gold_objects: dict[str, list[set]]) -> dict[str, dict]:
    num_groups_in_predicted_categories = list(map(len, predicted_objects.values()))
    num_groups_in_gold_categories = list(map(len, gold_objects.values()))

    if num_groups_in_gold_categories != num_groups_in_predicted_categories:
        raise ValueError('Predicted and gold category groups should match!')

    if predicted_objects.keys() != gold_objects.keys():
        raise ValueError('Predicted and gold categories should match!')

    if not len(num_groups_in_predicted_categories):
        raise ValueError('There should be at least one category!')

    num_groups = num_groups_in_predicted_categories[0]
    if any(ng != num_groups for ng in num_groups_in_predicted_categories):
        raise ValueError('Each category should have the same number of groups!')

    uncategorized_predicted_objects: list[set] = [set() for _ in range(num_groups)]
    uncategorized_gold_objects: list[set] = [set() for _ in range(num_groups)]

    for category in predicted_objects.keys():
        category_predicted_objects = predicted_objects[category]
        category_gold_objects = gold_objects[category]

        def categorize(obj) -> tuple:
            return category, obj

        for collected_group_objects, group_objects in zip(uncategorized_predicted_objects, category_predicted_objects):
            collected_group_objects.update(map(categorize, group_objects))
        for collected_group_objects, group_objects in zip(uncategorized_gold_objects, category_gold_objects):
            collected_group_objects.update(map(categorize, group_objects))

    macro_scores = _scores2dict(*ir_macro_scores(uncategorized_predicted_objects, uncategorized_gold_objects))
    micro_scores = _scores2dict(*ir_micro_scores(uncategorized_predicted_objects, uncategorized_gold_objects))
    final_scores = {**_with_prefix(macro_scores, 'macro'), **_with_prefix(micro_scores, 'micro'), 'categories': {}}

    for category in gold_objects.keys():
        macro_scores = _scores2dict(*ir_macro_scores(predicted_objects[category], gold_objects[category]))
        micro_scores = _scores2dict(*ir_micro_scores(predicted_objects[category], gold_objects[category]))
        final_scores['categories'][category] = {**_with_prefix(macro_scores, 'macro'), **_with_prefix(micro_scores, 'micro')}

    return final_scores



import asyncio
from argparse import ArgumentParser, Namespace
from typing import Callable

from talisman_tools.arguments.domain import get_domain_factory
from talisman_tools.arguments.reader import get_reader_factory
from tp_interfaces.abstract import AbstractDocumentProcessor


def configure_check_performance_parser(parser: ArgumentParser) -> None:
    domain_factory = get_domain_factory(parser)
    reader_factory = get_reader_factory(parser, {'docs_path': '<docs path>'})

    parser.add_argument('count', type=int, metavar='<launch count>', default=1)

    def get_action() -> Callable[[AbstractDocumentProcessor, Namespace], None]:
        def action(processor: AbstractDocumentProcessor, args: Namespace):
            from .action import async_measure

            domain_factory(args)
            docs = tuple(reader_factory(args)['docs_path'].read())

            asyncio.run(
                async_measure(docs, processor, args.count)
            )

        return action

    parser.set_defaults(processor_action=get_action)



from pathlib import Path
from typing import Callable, Dict

from talisman_tools.plugin.abstract import AbstractPluginManager, flattened
from tp_interfaces.abstract import ModelTypeFactory
from tp_interfaces.readers.abstract import AbstractReader


class EndpointPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'ENDPOINTS')

    @property
    def flattened(self) -> Dict[str, Callable[[Path], AbstractReader]]:
        return flattened(self.plugins, delimiter='/', start='/')

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, ModelTypeFactory)  # TODO: validate signature



from typing import Callable, Dict

from tp_interfaces.serializers.abstract import AbstractSerializer
from .abstract import AbstractPluginManager, flattened


class SerializerPluginManager(AbstractPluginManager):
    def __init__(self):
        AbstractPluginManager.__init__(self, 'SERIALIZERS')

    @property
    def flattened(self) -> Dict[str, Callable[[], AbstractSerializer]]:
        return flattened(self.plugins)

    @staticmethod
    def _check_value(value) -> bool:
        return isinstance(value, dict)  # TODO: improve validation



from argparse import Namespace

from talisman_tools.configure.configure import read_config
from tp_interfaces.abstract import AbstractDocumentProcessor
from tp_interfaces.readers.abstract import AbstractReader


async def async_action(processor: AbstractDocumentProcessor, serializer, reader: AbstractReader, args: Namespace):

    async with processor:
        processor_config_type = processor.config_type
        config = processor_config_type.model_validate(read_config(args.config)) if args.config else processor_config_type()
        serializer(processor.process_stream(reader.read(), config, args.batch))
